/*
 * Grdc - GTK+/Gnome Remote Desktop Client
 * Copyright (C) 2009 - Vic Lee 
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, 
 * Boston, MA 02111-1307, USA.
 */

#include <gtk/gtk.h>
#include <glib/gi18n.h>
#include "config.h"
#include "grdcpublic.h"
#include "grdcpref.h"
#include "grdcplug.h"

G_DEFINE_TYPE (GrdcPlug, grdc_plug, GTK_TYPE_EVENT_BOX)

enum {
    CONNECT_SIGNAL,
    DISCONNECT_SIGNAL,
    LAST_SIGNAL
};

typedef struct _GrdcPlugSignalData
{
    GrdcPlug *gp;
    const gchar *signal_name;
} GrdcPlugSignalData;

static guint grdc_plug_signals[LAST_SIGNAL] = { 0 };

static void
grdc_plug_class_init (GrdcPlugClass *klass)
{
    grdc_plug_signals[CONNECT_SIGNAL] =
        g_signal_new ("connect",
            G_TYPE_FROM_CLASS (klass),
            G_SIGNAL_RUN_LAST | G_SIGNAL_ACTION,
            G_STRUCT_OFFSET (GrdcPlugClass, connect),
            NULL, NULL,
            g_cclosure_marshal_VOID__VOID, G_TYPE_NONE, 0);
    grdc_plug_signals[DISCONNECT_SIGNAL] =
        g_signal_new ("disconnect",
            G_TYPE_FROM_CLASS (klass),
            G_SIGNAL_RUN_LAST | G_SIGNAL_ACTION,
            G_STRUCT_OFFSET (GrdcPlugClass, disconnect),
            NULL, NULL,
            g_cclosure_marshal_VOID__VOID, G_TYPE_NONE, 0);
}

static void
grdc_plug_init_cancel (GrdcInitDialog *dialog, gint response_id, GrdcPlug *gp)
{
    if ((response_id == GTK_RESPONSE_CANCEL || response_id == GTK_RESPONSE_DELETE_EVENT)
        && dialog->mode == GRDC_INIT_MODE_CONNECTING)
    {
        grdc_plug_close_connection (gp);
    }
}

static void
grdc_plug_show_init_dialog (GrdcPlug *gp, GtkWindow *parent, const gchar *name)
{
    if (gp->init_dialog)
    {
        gtk_widget_destroy (gp->init_dialog);
    }
    gp->init_dialog = grdc_init_dialog_new (parent, _("Connecting to '%s'..."), name);
    g_signal_connect (G_OBJECT (gp->init_dialog), "response", G_CALLBACK (grdc_plug_init_cancel), gp);
    gtk_widget_show (gp->init_dialog);
}

static void
grdc_plug_hide_init_dialog (GrdcPlug *gp)
{
    if (gp->init_dialog && GTK_IS_WIDGET (gp->init_dialog))
    {
        gtk_widget_destroy (gp->init_dialog);
    }
    gp->init_dialog = NULL;
}

static void
grdc_plug_destroy (GrdcPlug *gp, gpointer data)
{
    grdc_plug_hide_init_dialog (gp);
}

static void
grdc_plug_connect (GrdcPlug *gp, gpointer data)
{
    grdc_plug_hide_init_dialog (gp);
}

static void
grdc_plug_disconnect (GrdcPlug *gp, gpointer data)
{
    grdc_plug_hide_init_dialog (gp);
}

static void
grdc_plug_init (GrdcPlug *gp)
{
    g_signal_connect (G_OBJECT (gp), "destroy", G_CALLBACK (grdc_plug_destroy), NULL);
    g_signal_connect (G_OBJECT (gp), "connect", G_CALLBACK (grdc_plug_connect), NULL);
    g_signal_connect (G_OBJECT (gp), "disconnect", G_CALLBACK (grdc_plug_disconnect), NULL);

    gp->init_dialog = NULL;
    gp->grdc_file = NULL;
    gp->has_error = FALSE;
    gp->width = 0;
    gp->height = 0;
    gp->scale = FALSE;
    gp->error_message[0] = '\0';
    gp->ssh_tunnel = NULL;
}

gboolean
grdc_plug_open_connection (GrdcPlug *gp, GtkWindow *parent, GrdcFile *grdcfile)
{
    gp->grdc_file = grdcfile;

    /* Show "server" instead of "name" for quick connect */
    grdc_plug_show_init_dialog (gp, parent, (grdcfile->filename ? grdcfile->name : grdcfile->server));
    return (* (GRDC_PLUG_GET_CLASS (gp)->open_connection)) (gp);
}

gboolean
grdc_plug_close_connection (GrdcPlug *gp)
{
#ifdef HAVE_LIBSSH
    if (gp->ssh_tunnel)
    {
        grdc_ssh_tunnel_free (gp->ssh_tunnel);
        gp->ssh_tunnel = NULL;
    }
#endif
    return (* (GRDC_PLUG_GET_CLASS (gp)->close_connection)) (gp);
}

const gpointer
grdc_plug_query_feature (GrdcPlug *gp, GrdcPlugFeature feature)
{
    return (* (GRDC_PLUG_GET_CLASS (gp)->query_feature)) (gp, feature);
}

void
grdc_plug_call_feature (GrdcPlug *gp, GrdcPlugFeature feature, const gpointer data)
{
    (* (GRDC_PLUG_GET_CLASS (gp)->call_feature)) (gp, feature, data);
}

static gboolean
grdc_plug_emit_signal_timeout (gpointer user_data)
{
    GrdcPlugSignalData *data = (GrdcPlugSignalData*) user_data;
    g_signal_emit_by_name (G_OBJECT (data->gp), data->signal_name);
    g_free (data);
    return FALSE;
}

void
grdc_plug_emit_signal (GrdcPlug *gp, const gchar *signal_name)
{
    GrdcPlugSignalData *data;

    data = g_new (GrdcPlugSignalData, 1);
    data->gp = gp;
    data->signal_name = signal_name;
    TIMEOUT_ADD (0, grdc_plug_emit_signal_timeout, data);
}

gchar*
grdc_plug_start_tunnel (GrdcPlug *gp, gint default_port)
{
    gchar dest[200];
#ifdef HAVE_LIBSSH
    const gchar *sshuser;
    gchar sshserver[200], *s, *ptr;
    gchar *pubkey, *privkey;
    gint ret;
    GrdcSSHTunnel *tunnel;
#endif

    if (grdc_file_is_incoming (gp->grdc_file))
    {
        dest[0] = '\0';
    }
    else
    {
        if (g_strrstr (gp->grdc_file->server, ":") == NULL)
        {
            g_snprintf (dest, sizeof (dest), "%s:%i", gp->grdc_file->server, default_port);
        }
        else
        {
            g_strlcpy (dest, gp->grdc_file->server, sizeof (dest));
        }
    }

#ifdef HAVE_LIBSSH
    if (!gp->grdc_file->ssh_enabled || grdc_file_is_incoming (gp->grdc_file))
    {
        return g_strdup (dest);
    }

    /* Reuse existing SSH connection if it's reconnecting to destination */
    if (gp->ssh_tunnel == NULL)
    {

        if (gp->grdc_file->ssh_server && gp->grdc_file->ssh_server[0] != '\0')
        {
            if (g_strrstr (gp->grdc_file->ssh_server, ":") == NULL)
            {
                g_snprintf (sshserver, sizeof (sshserver), "%s:%i", gp->grdc_file->ssh_server, DEFAULT_SSH_PORT);
            }
            else
            {
                g_strlcpy (sshserver, gp->grdc_file->ssh_server, sizeof (sshserver));
            }
        }
        else if (gp->grdc_file->server == NULL || gp->grdc_file->server[0] == '\0')
        {
            return g_strdup (dest);
        }
        else
        {
            s = g_strdup (gp->grdc_file->server);
            ptr = g_strrstr (s, ":");
            if (ptr) *ptr = '\0';
            g_snprintf (sshserver, sizeof (sshserver), "%s:%i", s, DEFAULT_SSH_PORT);
            g_free (s);
        }

        THREADS_ENTER
        grdc_init_dialog_set_status (GRDC_INIT_DIALOG (gp->init_dialog),
            _("Connecting to SSH server %s..."), sshserver);
        THREADS_LEAVE

        sshuser = (gp->grdc_file->ssh_username && gp->grdc_file->ssh_username[0] != '\0' ?
                gp->grdc_file->ssh_username : g_get_user_name ());

        tunnel = grdc_ssh_tunnel_new (sshserver, sshuser);
        if (tunnel->error)
        {
            g_strlcpy (gp->error_message, tunnel->error, MAX_ERROR_LENGTH);
            grdc_ssh_tunnel_free (tunnel);
            gp->has_error = TRUE;
            return NULL;
        }

        switch (gp->grdc_file->ssh_auth)
        {

        case SSH_AUTH_PASSWORD:

            THREADS_ENTER
            grdc_init_dialog_set_status (GRDC_INIT_DIALOG (gp->init_dialog),
                _("Authenticating %s's password to SSH server %s..."), sshuser, sshserver);

            ret = grdc_init_dialog_authpwd (GRDC_INIT_DIALOG (gp->init_dialog), _("SSH Password"), FALSE);
            THREADS_LEAVE

            if (ret != GTK_RESPONSE_OK)
            {
                grdc_ssh_tunnel_free (tunnel);
                return NULL;
            }

            if (!grdc_ssh_tunnel_auth_password (tunnel, GRDC_INIT_DIALOG (gp->init_dialog)->password))
            {
                g_strlcpy (gp->error_message, tunnel->error, MAX_ERROR_LENGTH);
                grdc_ssh_tunnel_free (tunnel);
                gp->has_error = TRUE;
                return NULL;
            }

            break;

        case SSH_AUTH_PUBLICKEY:

            THREADS_ENTER
            grdc_init_dialog_set_status (GRDC_INIT_DIALOG (gp->init_dialog),
                _("Authenticating %s's identity to SSH server %s..."), sshuser, sshserver);
            THREADS_LEAVE

            s = (gp->grdc_file->ssh_privatekey && gp->grdc_file->ssh_privatekey[0] != '\0' ?
                g_strdup (gp->grdc_file->ssh_privatekey) : grdc_ssh_find_identity ());
            if (s == NULL)
            {
                g_strlcpy (gp->error_message, "Identity file not found", MAX_ERROR_LENGTH);
                grdc_ssh_tunnel_free (tunnel);
                gp->has_error = TRUE;
                return NULL;
            }
            privkey = grdc_ssh_identity_path (s);
            pubkey = g_strdup_printf ("%s.pub", privkey);
            g_free (s);

            ret = grdc_ssh_tunnel_auth_pubkey (tunnel, pubkey, privkey, "");
            if (ret == 0)
            {
                g_strlcpy (gp->error_message, tunnel->error, MAX_ERROR_LENGTH);
                grdc_ssh_tunnel_free (tunnel);
                g_free (privkey);
                g_free (pubkey);
                gp->has_error = TRUE;
                return NULL;
            }

            if (ret < 0)
            {
                THREADS_ENTER
                ret = grdc_init_dialog_authpwd (GRDC_INIT_DIALOG (gp->init_dialog), _("SSH Private Key Passphrase"), FALSE);
                THREADS_LEAVE

                if (ret != GTK_RESPONSE_OK)
                {
                    grdc_ssh_tunnel_free (tunnel);
                    g_free (privkey);
                    g_free (pubkey);
                    return NULL;
                }

                ret = grdc_ssh_tunnel_auth_pubkey (tunnel, pubkey, privkey, GRDC_INIT_DIALOG (gp->init_dialog)->password);
            }

            if (ret <= 0)
            {
                g_strlcpy (gp->error_message, tunnel->error, MAX_ERROR_LENGTH);
                grdc_ssh_tunnel_free (tunnel);
                g_free (privkey);
                g_free (pubkey);
                gp->has_error = TRUE;
                return NULL;
            }
            g_free (privkey);
            g_free (pubkey);
            break;

        default:
            return NULL;
        }

        gp->ssh_tunnel = tunnel;

    }

    THREADS_ENTER
    grdc_init_dialog_set_status (GRDC_INIT_DIALOG (gp->init_dialog),
        _("Connecting to %s through SSH tunnel..."), gp->grdc_file->server);
    THREADS_LEAVE

    if (!grdc_ssh_tunnel_start (gp->ssh_tunnel, dest, grdc_pref.sshtunnel_port))
    {
        g_strlcpy (gp->error_message, gp->ssh_tunnel->error, MAX_ERROR_LENGTH);
        grdc_ssh_tunnel_free (gp->ssh_tunnel);
        gp->has_error = TRUE;
        return NULL;
    }

    return g_strdup_printf ("127.0.0.1:%i", grdc_pref.sshtunnel_port);

#else

    return g_strdup (dest);

#endif
}


