/*
 * Grdc - GTK+/Gnome Remote Desktop Client
 * Copyright (C) 2009 - Vic Lee 
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, 
 * Boston, MA 02111-1307, USA.
 */

#include <gtk/gtk.h>
#include <glib/gstdio.h>
#include <glib/gi18n.h>
#include <stdlib.h>
#include <signal.h>
#include "config.h"
#ifdef HAVE_PTHREAD
#include <pthread.h>
#endif
#include "grdcpublic.h"
#include "grdcpref.h"
#include "grdcfile.h"
#include "grdcinitdialog.h"
#include "grdcplug.h"
#include "grdcplug_vnc.h"

G_DEFINE_TYPE (GrdcPlugVnc, grdc_plug_vnc, GRDC_TYPE_PLUG)

#define hidden_cursor_width 2
#define hidden_cursor_height 2
static const gchar hidden_cursor_bits[] = { 0x03, 0x03 };
static const gchar hidden_cursor_mask_bits[] = { 0x03, 0x03 };

static void
grdc_plug_vnc_scale_area (GrdcPlugVnc *gp_vnc, gint *x, gint *y, gint *w, gint *h)
{
    GrdcPlug *gp = GRDC_PLUG (gp_vnc);
    gint sx, sy, sw, sh;

    if (gp_vnc->rgb_buffer == NULL || gp_vnc->scale_buffer == NULL) return;

    /* We have to expend the scaled region one scaled pixel, to avoid gaps */
    sx = MIN (MAX (0, (*x) * gp_vnc->scale_width / gp->width
        - gp_vnc->scale_width / gp->width - 1), gp_vnc->scale_width - 1);
    sy = MIN (MAX (0, (*y) * gp_vnc->scale_height / gp->height
        - gp_vnc->scale_height / gp->height - 1), gp_vnc->scale_height - 1);
    sw = MIN (gp_vnc->scale_width - sx, (*w) * gp_vnc->scale_width / gp->width
        + gp_vnc->scale_width / gp->width + 1);
    sh = MIN (gp_vnc->scale_height - sy, (*h) * gp_vnc->scale_height / gp->height
        + gp_vnc->scale_height / gp->height + 1);

    gdk_pixbuf_scale (gp_vnc->rgb_buffer, gp_vnc->scale_buffer,
        sx, sy,
        sw, sh,
        0, 0,
        (double) gp_vnc->scale_width / (double) gp->width,
        (double) gp_vnc->scale_height / (double) gp->height,
        grdc_pref.scale_quality);

    *x = sx; *y = sy; *w = sw; *h = sh;
}

static gboolean
grdc_plug_vnc_update_scale_buffer (GrdcPlugVnc *gp_vnc)
{
    GrdcPlug *gp = GRDC_PLUG (gp_vnc);
    gint width, height;
    gint x, y, w, h;

    if (gp_vnc->running)
    {
        if (gp->scale)
        {
            width = GTK_WIDGET (gp_vnc)->allocation.width;
            height = GTK_WIDGET (gp_vnc)->allocation.height;
            if (width > 1 && height > 1)
            {
                if (gp_vnc->scale_buffer)
                {
                    g_object_unref (gp_vnc->scale_buffer);
                }
                gp_vnc->scale_buffer = gdk_pixbuf_new (GDK_COLORSPACE_RGB, TRUE, 8, width, height);
                gp_vnc->scale_width = width;
                gp_vnc->scale_height = height;

                x = 0; y = 0; w = gp->width; h = gp->height;
                grdc_plug_vnc_scale_area (gp_vnc, &x, &y, &w, &h);
                gtk_widget_queue_draw_area (GTK_WIDGET (gp_vnc), x, y, w, h);
            }
        }
        else
        {
            if (gp_vnc->scale_buffer)
            {
                g_object_unref (gp_vnc->scale_buffer);
                gp_vnc->scale_buffer = NULL;
            }
            gp_vnc->scale_width = 0;
            gp_vnc->scale_height = 0;
        }
    }
    gp_vnc->scale_handler = 0;
    return FALSE;
}

static void
grdc_plug_vnc_update_scale (GrdcPlugVnc *gp_vnc, gboolean scale)
{
    GrdcPlug *gp = GRDC_PLUG (gp_vnc);
    gp->scale = scale;
    if (scale)
    {
        gtk_widget_set_size_request (GTK_WIDGET (gp_vnc->drawing_area), -1, -1);
    }
    else
    {
        gtk_widget_set_size_request (GTK_WIDGET (gp_vnc->drawing_area), gp->width, gp->height);
    }
}

#ifdef HAVE_LIBVNCCLIENT

typedef struct _GrdcKeyVal
{
    guint keyval;
    guint16 keycode;
} GrdcKeyVal;

/***************************** LibVNCClient related codes *********************************/
#include <rfb/rfbclient.h>

// same as SupportsClient2Server
#define rfbSupportsClient2Server(client,messageType) \
    ((client)->supportedMessages.client2server[((messageType & 0xFF)/8)] & (1<<(messageType % 8)))

typedef struct _GrdcPlugVncParam
{
    rfbClient *cl;
    gint x, y, w, h;
} GrdcPlugVncParam;

static void
grdc_plug_vnc_update_quality (rfbClient *cl, gint quality)
{
    switch (quality)
    {
    case 9:
        cl->appData.useBGR233 = 0;
        cl->appData.encodingsString = "copyrect hextile raw";
        cl->appData.compressLevel = 0;
        cl->appData.qualityLevel = 9;
        break;
    case 2:
        cl->appData.useBGR233 = 0;
        cl->appData.encodingsString = "tight zrle ultra copyrect hextile zlib corre rre raw";
        cl->appData.compressLevel = 3;
        cl->appData.qualityLevel = 7;
        break;
    case 1:
        cl->appData.useBGR233 = 0;
        cl->appData.encodingsString = "tight zrle ultra copyrect hextile zlib corre rre raw";
        cl->appData.compressLevel = 5;
        cl->appData.qualityLevel = 5;
        break;
    case 0:
    default:
        cl->appData.useBGR233 = 1;
        cl->appData.encodingsString = "tight zrle ultra copyrect hextile zlib corre rre raw";
        cl->appData.compressLevel = 9;
        cl->appData.qualityLevel = 0;
        break;
    }
}

static void
grdc_plug_vnc_update_colordepth (rfbClient *cl, gint colordepth)
{
    cl->format.depth = colordepth;
    cl->appData.requestedDepth = colordepth;

    switch (colordepth)
    {
    case 24:
        cl->format.bitsPerPixel = 32;
        cl->format.redMax = 0xff;
        cl->format.greenMax = 0xff;
        cl->format.blueMax = 0xff;
        cl->format.redShift = 0;
        cl->format.greenShift = 8;
        cl->format.blueShift = 16;
        break;
    case 16:
        cl->format.bitsPerPixel = 16;
        cl->format.redMax = 0x3f;
        cl->format.greenMax = 0x1f;
        cl->format.blueMax = 0x1f;
        cl->format.redShift = 0;
        cl->format.greenShift = 6;
        cl->format.blueShift = 11;
        break;
    case 15:
        cl->format.bitsPerPixel = 16;
        cl->format.redMax = 0x1f;
        cl->format.greenMax = 0x1f;
        cl->format.blueMax = 0x1f;
        cl->format.redShift = 0;
        cl->format.greenShift = 5;
        cl->format.blueShift = 10;
        break;
    case 8:
    default:
        cl->format.bitsPerPixel = 8;
        cl->format.redMax = 7;
        cl->format.greenMax = 7;
        cl->format.blueMax = 3;
        cl->format.redShift = 0;
        cl->format.greenShift = 3;
        cl->format.blueShift = 6;
        break;
    }
}

static rfbBool
grdc_plug_vnc_rfb_allocfb (rfbClient *cl)
{
    GrdcPlugVnc *gp_vnc;
    gint width, height, depth, size;

    gp_vnc = GRDC_PLUG_VNC (rfbClientGetClientData (cl, NULL));

    width = cl->width;
    height = cl->height;
    depth = cl->format.bitsPerPixel;
    size = width * height * (depth / 8);

    if (gp_vnc->rgb_buffer) g_object_unref (gp_vnc->rgb_buffer);
    if (gp_vnc->vnc_buffer) g_free (gp_vnc->vnc_buffer);
    
    gp_vnc->rgb_buffer = gdk_pixbuf_new (GDK_COLORSPACE_RGB, TRUE, 8, width, height);
    gdk_pixbuf_fill (gp_vnc->rgb_buffer, 0);

    if (depth == 32)
    {
        gp_vnc->vnc_buffer = NULL;
        cl->frameBuffer = gdk_pixbuf_get_pixels (gp_vnc->rgb_buffer);
    }
    else
    {
        gp_vnc->vnc_buffer = (guchar*) g_malloc (size);
        cl->frameBuffer = gp_vnc->vnc_buffer;
    }

    return TRUE;
}

static gint
grdc_plug_vnc_bits (gint n)
{
    gint b = 0;
    while (n) { b++; n >>= 1; }
    return b;
}

static gboolean
grdc_plug_vnc_queue_draw_area (GrdcPlugVncParam *param)
{
    GrdcPlugVnc *gp_vnc;

    gp_vnc = GRDC_PLUG_VNC (rfbClientGetClientData (param->cl, NULL));
    if (!gp_vnc->closed && gp_vnc->connected)
    {
        gtk_widget_queue_draw_area (GTK_WIDGET (gp_vnc), param->x, param->y, param->w, param->h);
    }
    g_free (param);
    return FALSE;
}

static void
grdc_plug_vnc_rfb_updatefb (rfbClient* cl, int x, int y, int w, int h)
{
    GrdcPlugVnc *gp_vnc;
    GrdcPlug *gp;
    gint ix, iy, x2, y2;
    guchar *destptr, *srcptr;
    gint bytesPerPixel;
    gint i;
    guint32 pixel;
    gint rs, gs, bs, rm, gm, bm, rb, gb, bb;
    GrdcPlugVncParam *param;

    gp_vnc = GRDC_PLUG_VNC (rfbClientGetClientData (cl, NULL));
    gp = GRDC_PLUG (gp_vnc);

    switch (cl->format.bitsPerPixel)
    {
    case 32:
        /* The following codes fill in the Alpha channel, otherwise the GdkPixbuf scaling won't work */
        x2 = x + w;
        y2 = y + h;
        for (iy = y; iy < y2; iy++)
        {
            destptr = gdk_pixbuf_get_pixels (gp_vnc->rgb_buffer) + ((iy * gp->width + x) * 4) + 3;
            for (ix = x; ix < x2; ix++)
            {
                *destptr = 0xff;
                destptr += 4;
            }
        }
        break;
    default:
        if (w < 1 || h < 1) break;
        x2 = x + w;
        y2 = y + h;
        bytesPerPixel = cl->format.bitsPerPixel / 8;
        rm = cl->format.redMax;
        gm = cl->format.greenMax;
        bm = cl->format.blueMax;
        rb = 8 - grdc_plug_vnc_bits (rm);
        gb = 8 - grdc_plug_vnc_bits (gm);
        bb = 8 - grdc_plug_vnc_bits (bm);
        rs = cl->format.redShift;
        gs = cl->format.greenShift;
        bs = cl->format.blueShift;
        for (iy = y; iy < y2; iy++)
        {
            destptr = gdk_pixbuf_get_pixels (gp_vnc->rgb_buffer) + ((iy * gp->width + x) * 4);
            srcptr = gp_vnc->vnc_buffer + ((iy * gp->width + x) * bytesPerPixel);
            for (ix = x; ix < x2; ix++)
            {
                pixel = 0;
                for (i = 0; i < bytesPerPixel; i++) pixel += (*srcptr++) << (8 * i);
                *destptr++ = ((pixel >> rs) & rm) << rb;
                *destptr++ = ((pixel >> gs) & gm) << gb;
                *destptr++ = ((pixel >> bs) & bm) << bb;
                *destptr++ = 0xff;
            }
        }
        break;
    }

    if (gp->scale)
    {
        grdc_plug_vnc_scale_area (gp_vnc, &x, &y, &w, &h);
    }

    param = g_new (GrdcPlugVncParam, 1);
    param->cl = cl;
    param->x = x;
    param->y = y;
    param->w = w;
    param->h = h;
    IDLE_ADD ((GSourceFunc) grdc_plug_vnc_queue_draw_area, param);
}

static void
grdc_plug_vnc_rfb_cuttext (rfbClient* cl, const char *text, int textlen)
{
    GrdcPlugVnc *gp_vnc;
    GTimeVal t;
    glong diff;

    gp_vnc = GRDC_PLUG_VNC (rfbClientGetClientData (cl, NULL));
    g_get_current_time (&t);
    diff = (t.tv_sec - gp_vnc->clipboard_timer.tv_sec) * 10 + (t.tv_usec - gp_vnc->clipboard_timer.tv_usec) / 100000;
    if (diff < 10) return;

    gp_vnc->clipboard_timer = t;

    THREADS_ENTER
    gtk_clipboard_set_text (gtk_clipboard_get (GDK_SELECTION_CLIPBOARD), text, textlen);
    THREADS_LEAVE
}

static char*
grdc_plug_vnc_rfb_password (rfbClient *cl)
{
    GrdcPlugVnc *gp_vnc;
    GrdcPlug *gp;
    GtkWidget *dialog;
    gint ret;
    gchar *pwd = NULL;

    gp_vnc = GRDC_PLUG_VNC (rfbClientGetClientData (cl, NULL));
    gp_vnc->auth_called = TRUE;
    gp = GRDC_PLUG (gp_vnc);

    dialog = gp->init_dialog;

    if (gp->grdc_file->password && gp->grdc_file->password[0] != '\0')
    {
        pwd = g_strdup (gp->grdc_file->password);
    }
    else
    {
        THREADS_ENTER
        ret = grdc_init_dialog_authpwd (GRDC_INIT_DIALOG (dialog), _("VNC Password"), 
            (gp->grdc_file->filename != NULL));
        THREADS_LEAVE

        if (ret == GTK_RESPONSE_OK)
        {
            pwd = g_strdup (GRDC_INIT_DIALOG (dialog)->password);
        }
        else
        {
            gp_vnc->connected = FALSE;
        }
    }
    return pwd;
}

/* Translate known VNC messages. It's for intltool only, not for gcc */
#ifdef __DO_NOT_COMPILE_ME__
N_("Unable to connect to VNC server")
N_("Couldn't convert '%s' to host address")
N_("VNC connection failed: %s")
N_("Your connection has been rejected.")
#endif
/* TODO: We only store the last message at this moment. */
static gchar vnc_error[MAX_ERROR_LENGTH + 1];
static void
grdc_plug_vnc_rfb_output(const char *format, ...)
{
    va_list args;
    va_start(args, format);
    gchar *f, *p;

    /* eliminate the last \n */
    f = g_strdup (format);
    if (f[strlen (f) - 1] == '\n') f[strlen (f) - 1] = '\0';

/*g_printf("%s,len=%i\n", f, strlen(f));*/
    if (g_strcmp0 (f, "VNC connection failed: %s") == 0)
    {
        p = va_arg (args, gchar*);
/*g_printf("(param)%s,len=%i\n", p, strlen(p));*/
        g_snprintf (vnc_error, MAX_ERROR_LENGTH, _(f), _(p));
    }
    else
    {
        g_vsnprintf (vnc_error, MAX_ERROR_LENGTH, _(f), args);
    }

    g_free (f);

    va_end(args);
}

static gboolean
grdc_plug_vnc_incoming_connection (GrdcPlugVnc *gp_vnc, rfbClient *cl)
{
    GrdcPlug *gp = GRDC_PLUG (gp_vnc);
    fd_set fds;

    gp_vnc->listen_sock = ListenAtTcpPort (cl->listenPort);
    if (gp_vnc->listen_sock < 0) return FALSE;

    grdc_init_dialog_set_status (GRDC_INIT_DIALOG (gp->init_dialog),
        _("Listening on Port %i for an Incoming VNC Connection..."), cl->listenPort);

    FD_ZERO (&fds); 
    FD_SET (gp_vnc->listen_sock, &fds);
    select (FD_SETSIZE, &fds, NULL, NULL, NULL);

    if (!FD_ISSET (gp_vnc->listen_sock, &fds))
    {
        close (gp_vnc->listen_sock);
        gp_vnc->listen_sock = -1;
        return FALSE;
    }

    cl->sock = AcceptTcpConnection (gp_vnc->listen_sock);
    close (gp_vnc->listen_sock);
    gp_vnc->listen_sock = -1;
    if (cl->sock < 0 || !SetNonBlocking (cl->sock))
    {
        return FALSE;
    }

    return TRUE;
}

static gboolean
grdc_plug_vnc_main_loop (GrdcPlugVnc *gp_vnc)
{
    gint ret;
    rfbClient *cl;

    if (!gp_vnc->connected)
    {
        gp_vnc->running = FALSE;
        return FALSE;
    }

    cl = (rfbClient*) gp_vnc->client;

/* We need to wait for a while in a thread, to avoid 100% CPU usage in a loop without sleep */
#ifdef HAVE_PTHREAD
    ret = WaitForMessage (cl, 100);
#else
    ret = WaitForMessage (cl, 0);
#endif
    if (ret < 0)
    {
        gp_vnc->running = FALSE;
        return FALSE;
    }
    if (ret == 0) return TRUE;

    if (!HandleRFBServerMessage (cl))
    {
        gp_vnc->running = FALSE;
        return FALSE;
    }

    return TRUE;
}

static gboolean
grdc_plug_vnc_main (GrdcPlugVnc *gp_vnc)
{
    GrdcPlug *gp = GRDC_PLUG (gp_vnc);
    rfbClient *cl = NULL;
    gchar *host, *pos;

    gp_vnc->running = TRUE;

    rfbClientLog = grdc_plug_vnc_rfb_output;
    rfbClientErr = grdc_plug_vnc_rfb_output;

    while (gp_vnc->connected)
    {
        gp_vnc->auth_called = FALSE;

        host = grdc_plug_start_tunnel (gp, 5900);

        if (host == NULL)
        {
            gp_vnc->connected = FALSE;
            break;
        }

        cl = rfbGetClient(8, 3, 4);
        cl->MallocFrameBuffer = grdc_plug_vnc_rfb_allocfb;
        cl->canHandleNewFBSize = TRUE;
        cl->GetPassword = grdc_plug_vnc_rfb_password;
        cl->GotFrameBufferUpdate = grdc_plug_vnc_rfb_updatefb;
        cl->GotXCutText = grdc_plug_vnc_rfb_cuttext;
        rfbClientSetClientData (cl, NULL, gp_vnc);

        if (host[0] == '\0')
        {
            cl->serverHost = host;
            cl->listenSpecified = TRUE;
            cl->listenPort = gp->grdc_file->listenport;

            grdc_plug_vnc_incoming_connection (gp_vnc, cl);
        }
        else
        {
            pos = g_strrstr (host, ":");
            *pos++ = '\0';
            cl->serverPort = MAX (0, atoi (pos));
            cl->serverHost = host;

            /* Support short-form (:0, :1) */
            if (cl->serverPort < 100) cl->serverPort += 5900;
        }

        cl->appData.useRemoteCursor = (gp->grdc_file->showcursor ? FALSE : TRUE);

        grdc_plug_vnc_update_quality (cl, gp->grdc_file->quality);
        grdc_plug_vnc_update_colordepth (cl, gp->grdc_file->colordepth);
        SetFormatAndEncodings (cl);

        if (rfbInitClient (cl, NULL, NULL)) break;

        /* If the authentication is not called, it has to be a fatel error and must quit */
        if (!gp_vnc->auth_called)
        {
            gp_vnc->connected = FALSE;
            break;
        }

        /* Otherwise, it's a password error. Try to clear saved password if any */
        if (gp->grdc_file->password) gp->grdc_file->password[0] = '\0';

        if (!gp_vnc->connected) break;

        THREADS_ENTER
        grdc_init_dialog_set_status_temp (GRDC_INIT_DIALOG (gp->init_dialog),
            _("VNC authentication failed. Trying to reconnect..."));
        THREADS_LEAVE
        /* It's safer to sleep a while before reconnect */
        sleep (2);
    }

    if (!gp_vnc->connected)
    {
        if (cl && !gp_vnc->auth_called && !(gp->has_error))
        {
            g_snprintf (gp->error_message, MAX_ERROR_LENGTH, "%s", vnc_error);
            gp->has_error = TRUE;
        }
        gp_vnc->running = FALSE;

        IDLE_ADD ((GSourceFunc) grdc_plug_close_connection, gp_vnc);

        return FALSE;
    }

    /* Save the password if it's requested to do so */
    if (GRDC_INIT_DIALOG (gp->init_dialog)->save_password)
    {
        if (gp->grdc_file->password) g_free (gp->grdc_file->password);
        gp->grdc_file->password =
            g_strdup (GRDC_INIT_DIALOG (gp->init_dialog)->password);
        grdc_file_save (gp->grdc_file);
    }

    gp_vnc->client = cl;
    gp->width = cl->width;
    gp->height = cl->height;

    THREADS_ENTER
    grdc_plug_vnc_update_scale (gp_vnc, gp->scale);
    grdc_plug_vnc_update_scale_buffer (gp_vnc);
    THREADS_LEAVE

    grdc_plug_emit_signal (gp, "connect");

    if (gp->grdc_file->disableserverinput)
    {
        PermitServerInput(cl, 1);
    }

    if (gp_vnc->thread)
    {
        while (grdc_plug_vnc_main_loop (gp_vnc)) { }
        gp_vnc->running = FALSE;
    }
    else
    {
        IDLE_ADD ((GSourceFunc) grdc_plug_vnc_main_loop, gp_vnc);
    }

    return FALSE;
}

#ifdef HAVE_PTHREAD
static gpointer
grdc_plug_vnc_main_thread (gpointer data)
{
    pthread_setcancelstate (PTHREAD_CANCEL_ENABLE, NULL);

    CANCEL_ASYNC
    grdc_plug_vnc_main (GRDC_PLUG_VNC (data));
    return NULL;
}
#endif

static gboolean
grdc_plug_vnc_on_motion (GtkWidget *widget, GdkEventMotion *event, GrdcPlugVnc *gp_vnc)
{
    GrdcPlug *gp = GRDC_PLUG (gp_vnc);
    gint x, y;

    if (!gp_vnc->connected || !gp_vnc->client) return FALSE;
    if (gp->grdc_file->viewonly) return FALSE;

    if (gp->scale)
    {
        x = event->x * gp->width / gp_vnc->scale_width;
        y = event->y * gp->height / gp_vnc->scale_height;
    }
    else
    {
        x = event->x;
        y = event->y;
    }
    SendPointerEvent ((rfbClient*) gp_vnc->client, x, y, gp_vnc->button_mask);
    return TRUE;
}

static gboolean
grdc_plug_vnc_on_button (GtkWidget *widget, GdkEventButton *event, GrdcPlugVnc *gp_vnc)
{
    GrdcPlug *gp = GRDC_PLUG (gp_vnc);
    gint x, y;
    gint mask;

    if (!gp_vnc->connected || !gp_vnc->client) return FALSE;
    if (gp->grdc_file->viewonly) return FALSE;

    /* We only accept 3 buttons */
    if (event->button < 1 || event->button > 3) return FALSE;
    /* We bypass 2button-press and 3button-press events */
    if (event->type != GDK_BUTTON_PRESS && event->type != GDK_BUTTON_RELEASE) return TRUE;

    mask = (1 << (event->button - 1));
    gp_vnc->button_mask = (event->type == GDK_BUTTON_PRESS ?
        (gp_vnc->button_mask | mask) :
        (gp_vnc->button_mask & (0xff - mask)));
    if (gp->scale)
    {
        x = event->x * gp->width / gp_vnc->scale_width;
        y = event->y * gp->height / gp_vnc->scale_height;
    }
    else
    {
        x = event->x;
        y = event->y;
    }
    SendPointerEvent ((rfbClient*) gp_vnc->client, x, y, gp_vnc->button_mask);
    return TRUE;
}

static gboolean
grdc_plug_vnc_on_scroll (GtkWidget *widget, GdkEventScroll *event, GrdcPlugVnc *gp_vnc)
{
    GrdcPlug *gp = GRDC_PLUG (gp_vnc);
    gint x, y;
    gint mask;

    if (!gp_vnc->connected || !gp_vnc->client) return FALSE;
    if (gp->grdc_file->viewonly) return FALSE;

    switch (event->direction)
    {
    case GDK_SCROLL_UP:
        mask = (1 << 3);
        break;
    case GDK_SCROLL_DOWN:
        mask = (1 << 4);
        break;
    case GDK_SCROLL_LEFT:
        mask = (1 << 5);
        break;
    case GDK_SCROLL_RIGHT:
        mask = (1 << 6);
        break;
    default:
        return FALSE;
    }

    if (gp->scale)
    {
        x = event->x * gp->width / gp_vnc->scale_width;
        y = event->y * gp->height / gp_vnc->scale_height;
    }
    else
    {
        x = event->x;
        y = event->y;
    }
    SendPointerEvent ((rfbClient*) gp_vnc->client, x, y, mask | gp_vnc->button_mask);
    SendPointerEvent ((rfbClient*) gp_vnc->client, x, y, gp_vnc->button_mask);

    return TRUE;
}

static void
grdc_plug_vnc_release_key (GrdcPlugVnc *gp_vnc, guint16 keycode)
{
    GrdcKeyVal *k;
    gint i;

    if (keycode == 0)
    {
        /* Send all release key events for previously pressed keys */
        for (i = 0; i < gp_vnc->pressed_keys->len; i++)
        {
            k = g_ptr_array_index (gp_vnc->pressed_keys, i);
            SendKeyEvent ((rfbClient*) gp_vnc->client, k->keyval, FALSE);
            g_free (k);
        }
        g_ptr_array_set_size (gp_vnc->pressed_keys, 0);
    }
    else
    {
        /* Unregister the keycode only */
        for (i = 0; i < gp_vnc->pressed_keys->len; i++)
        {
            k = g_ptr_array_index (gp_vnc->pressed_keys, i);
            if (k->keycode == keycode)
            {
                g_free (k);
                g_ptr_array_remove_index_fast (gp_vnc->pressed_keys, i);
                break;
            }
        }
    }
}

static gboolean
grdc_plug_vnc_on_key (GtkWidget *widget, GdkEventKey *event, GrdcPlugVnc *gp_vnc)
{
    GrdcKeyVal *k;

    if (!gp_vnc->connected || !gp_vnc->client) return FALSE;
    if (GRDC_PLUG (gp_vnc)->grdc_file->viewonly) return FALSE;

    /* Lucky! GDK uses the same XKeySym as rfbClient. I haven't yet found any key (or key-combo) not working */
    SendKeyEvent ((rfbClient*) gp_vnc->client, event->keyval, (event->type == GDK_KEY_PRESS ? TRUE : FALSE));

    /* Register/unregister the pressed key */
    if (event->type == GDK_KEY_PRESS)
    {
        k = g_new (GrdcKeyVal, 1);
        k->keyval = event->keyval;
        k->keycode = event->hardware_keycode;
        g_ptr_array_add (gp_vnc->pressed_keys, k);
    }
    else
    {
        grdc_plug_vnc_release_key (gp_vnc, event->hardware_keycode);
    }
    return TRUE;
}

static void
grdc_plug_vnc_on_cuttext_request (GtkClipboard *clipboard, const gchar *text, GrdcPlugVnc *gp_vnc)
{
    GTimeVal t;
    glong diff;

    if (text)
    {
        /* A timer (1 second) to avoid clipboard "loopback": text cut out from VNC won't paste back into VNC */
        g_get_current_time (&t);
        diff = (t.tv_sec - gp_vnc->clipboard_timer.tv_sec) * 10 + (t.tv_usec - gp_vnc->clipboard_timer.tv_usec) / 100000;
        if (diff < 10) return;

        gp_vnc->clipboard_timer = t;
        SendClientCutText ((rfbClient*) gp_vnc->client, (char*) text, strlen(text));
    }
}

static void
grdc_plug_vnc_on_cuttext (GtkClipboard *clipboard, GdkEvent *event, GrdcPlugVnc *gp_vnc)
{
    if (!gp_vnc->connected || !gp_vnc->client) return;
    if (GRDC_PLUG (gp_vnc)->grdc_file->viewonly) return;

    gtk_clipboard_request_text (clipboard, (GtkClipboardTextReceivedFunc) grdc_plug_vnc_on_cuttext_request, gp_vnc);
}

static void
grdc_plug_vnc_on_realize (GrdcPlugVnc *gp_vnc, gpointer data)
{
    GrdcPlug *gp = GRDC_PLUG (gp_vnc);
    GdkCursor *cursor;
    GdkPixmap *source, *mask;
    GdkColor fg, bg;

    if (gp->grdc_file->showcursor)
    {
        /* Hide local cursor (a small black dot) */
        gdk_color_parse ("black", &fg);
        gdk_color_parse ("white", &bg);
        source = gdk_bitmap_create_from_data (NULL, hidden_cursor_bits, hidden_cursor_width, hidden_cursor_height);
        mask = gdk_bitmap_create_from_data (NULL, hidden_cursor_mask_bits, hidden_cursor_width, hidden_cursor_height);
        cursor = gdk_cursor_new_from_pixmap (source, mask, &fg, &bg, 0, 0);
        gdk_pixmap_unref (source);
        gdk_pixmap_unref (mask);
        gdk_window_set_cursor (GTK_WIDGET (gp)->window, cursor);
        gdk_cursor_unref (cursor);
    }
}

/******************************************************************************************/

static gboolean
grdc_plug_vnc_open_connection (GrdcPlug *gp)
{
    GrdcPlugVnc *gp_vnc = GRDC_PLUG_VNC (gp);

    gp_vnc->connected = TRUE;

    g_signal_connect (G_OBJECT (gp_vnc), "realize",
        G_CALLBACK (grdc_plug_vnc_on_realize), NULL);
    g_signal_connect (G_OBJECT (gp_vnc->drawing_area), "motion-notify-event",
        G_CALLBACK (grdc_plug_vnc_on_motion), gp_vnc);
    g_signal_connect (G_OBJECT (gp_vnc->drawing_area), "button-press-event",
        G_CALLBACK (grdc_plug_vnc_on_button), gp_vnc);
    g_signal_connect (G_OBJECT (gp_vnc->drawing_area), "button-release-event",
        G_CALLBACK (grdc_plug_vnc_on_button), gp_vnc);
    g_signal_connect (G_OBJECT (gp_vnc->drawing_area), "scroll-event",
        G_CALLBACK (grdc_plug_vnc_on_scroll), gp_vnc);
    g_signal_connect (G_OBJECT (gp_vnc->drawing_area), "key-press-event",
        G_CALLBACK (grdc_plug_vnc_on_key), gp_vnc);
    g_signal_connect (G_OBJECT (gp_vnc->drawing_area), "key-release-event",
        G_CALLBACK (grdc_plug_vnc_on_key), gp_vnc);

    gp_vnc->clipboard_handler = g_signal_connect (G_OBJECT (gtk_clipboard_get (GDK_SELECTION_CLIPBOARD)),
        "owner-change", G_CALLBACK (grdc_plug_vnc_on_cuttext), gp_vnc);

#ifdef HAVE_PTHREAD
    if (pthread_create (&gp_vnc->thread, NULL, grdc_plug_vnc_main_thread, gp_vnc))
    {
        /* I don't think this will ever happen... */
        g_print ("Failed to initialize pthread. Falling back to non-thread mode...\n");
        g_timeout_add (0, (GSourceFunc) grdc_plug_vnc_main, gp);
        gp_vnc->thread = 0;
    }
#else
    g_timeout_add (0, (GSourceFunc) grdc_plug_vnc_main, gp);
#endif

    return TRUE;
}

static gboolean
grdc_plug_vnc_close_connection_timeout (GrdcPlugVnc *gp_vnc)
{
    gint sock;

    /* wait until the running attribute is set to false by the VNC thread */
    if (gp_vnc->running) return TRUE;

    /* unregister the clipboard monitor */
    if (gp_vnc->clipboard_handler)
    {
        g_signal_handler_disconnect (G_OBJECT (gtk_clipboard_get (GDK_SELECTION_CLIPBOARD)),
            gp_vnc->clipboard_handler);
        gp_vnc->clipboard_handler = 0;
    }

    if (gp_vnc->scale_handler)
    {
        g_source_remove (gp_vnc->scale_handler);
        gp_vnc->scale_handler = 0;
    }
    if (gp_vnc->listen_sock >= 0)
    {
        close (gp_vnc->listen_sock);
    }
    if (gp_vnc->client)
    {
        sock = ((rfbClient*) gp_vnc->client)->sock;
        if (sock >= 0)
        {
            close (sock);
        }
        rfbClientCleanup((rfbClient*) gp_vnc->client);
        gp_vnc->client = NULL;
    }
    if (gp_vnc->rgb_buffer)
    {
        g_object_unref (gp_vnc->rgb_buffer);
        gp_vnc->rgb_buffer = NULL;
    }
    if (gp_vnc->vnc_buffer)
    {
        g_free (gp_vnc->vnc_buffer);
        gp_vnc->vnc_buffer = NULL;
    }
    if (gp_vnc->scale_buffer)
    {
        g_object_unref (gp_vnc->scale_buffer);
        gp_vnc->scale_buffer = NULL;
    }
    g_ptr_array_free (gp_vnc->pressed_keys, TRUE);

    grdc_plug_emit_signal (GRDC_PLUG (gp_vnc), "disconnect");

    return FALSE;
}

static gboolean
grdc_plug_vnc_close_connection (GrdcPlug *gp)
{
    GrdcPlugVnc *gp_vnc = GRDC_PLUG_VNC (gp);

    if (gp_vnc->closed) return FALSE;
    gp_vnc->closed = TRUE;
    gp_vnc->connected = FALSE;

#ifdef HAVE_PTHREAD
    if (gp_vnc->thread)
    {
        pthread_cancel (gp_vnc->thread);
        pthread_join (gp_vnc->thread, NULL);
        gp_vnc->running = FALSE;
        grdc_plug_vnc_close_connection_timeout (gp_vnc);
    }
    else
    {
        g_timeout_add (200, (GSourceFunc) grdc_plug_vnc_close_connection_timeout, gp);
    }
#else
    g_timeout_add (200, (GSourceFunc) grdc_plug_vnc_close_connection_timeout, gp);
#endif

    return FALSE;
}

static const gpointer
grdc_plug_vnc_query_feature (GrdcPlug *gp, GrdcPlugFeature feature)
{
    switch (feature)
    {
        case GRDC_PLUG_FEATURE_PREF:
        case GRDC_PLUG_FEATURE_PREF_QUALITY:
        case GRDC_PLUG_FEATURE_PREF_VIEWONLY:
        case GRDC_PLUG_FEATURE_UNFOCUS:
        case GRDC_PLUG_FEATURE_SCALE:
            return GINT_TO_POINTER (1);
        case GRDC_PLUG_FEATURE_PREF_DISABLESERVERINPUT:
            return (rfbSupportsClient2Server ((rfbClient*) (GRDC_PLUG_VNC (gp)->client), rfbSetServerInput) ?
                GINT_TO_POINTER (1) : NULL);
        default:
            return NULL;
    }
}

static void
grdc_plug_vnc_call_feature (GrdcPlug *gp, GrdcPlugFeature feature, const gpointer data)
{
    switch (feature)
    {
        case GRDC_PLUG_FEATURE_PREF_QUALITY:
            grdc_plug_vnc_update_quality ((rfbClient*) (GRDC_PLUG_VNC (gp)->client), GPOINTER_TO_INT (data));
            SetFormatAndEncodings ((rfbClient*) (GRDC_PLUG_VNC (gp)->client));
            break;
        case GRDC_PLUG_FEATURE_PREF_VIEWONLY:
            gp->grdc_file->viewonly = (data != NULL);
            break;
        case GRDC_PLUG_FEATURE_PREF_DISABLESERVERINPUT:
            PermitServerInput ((rfbClient*) (GRDC_PLUG_VNC (gp)->client), (data ? 1 : 0));
            gp->grdc_file->disableserverinput = (data ? TRUE : FALSE);
            break;
        case GRDC_PLUG_FEATURE_UNFOCUS:
            grdc_plug_vnc_release_key (GRDC_PLUG_VNC (gp), 0);
            break;
        case GRDC_PLUG_FEATURE_SCALE:
            grdc_plug_vnc_update_scale (GRDC_PLUG_VNC (gp), (data != NULL));
            break;
        default:
            break;
    }
}

#else /* HAVE_LIBVNCCLIENT */

static gboolean
grdc_plug_vnc_open_connection (GrdcPlug *gp)
{
    GtkWidget *dialog;
    /* This should never happen because if no VNC support users are not able to select VNC in preference dialog */
    dialog = gtk_message_dialog_new (NULL,
        GTK_DIALOG_MODAL, GTK_MESSAGE_ERROR, GTK_BUTTONS_CLOSE,
        "VNC not supported");
    gtk_dialog_run (GTK_DIALOG (dialog));
    gtk_widget_destroy (dialog);
    return FALSE;
}

static gboolean
grdc_plug_vnc_close_connection (GrdcPlug *gp)
{
    return FALSE;
}

static const gpointer
grdc_plug_vnc_query_feature (GrdcPlug *gp, GrdcPlugFeature feature)
{
    return NULL;
}

static void
grdc_plug_vnc_call_feature (GrdcPlug *gp, GrdcPlugFeature feature, const gpointer data)
{
}

#endif /* HAVE_LIBVNCCLIENT */

static gboolean
grdc_plug_vnc_on_expose (GtkWidget *widget, GdkEventExpose *event, GrdcPlugVnc *gp_vnc)
{
    GrdcPlug *gp = GRDC_PLUG (gp_vnc);
    GdkPixbuf *buffer;
    gint width, height, x, y;

    /* widget == gp_vnc->drawing_area */
    buffer = (gp->scale ? gp_vnc->scale_buffer : gp_vnc->rgb_buffer);
    if (!buffer) return FALSE;

    width = (gp->scale ? gp_vnc->scale_width : gp->width);
    height = (gp->scale ? gp_vnc->scale_height : gp->height);
    if (event->area.x >= width || event->area.y >= height) return FALSE;
    x = event->area.x;
    y = event->area.y;

    /* this is a little tricky. It "moves" the rgb_buffer pointer to (x,y) as top-left corner,
       and keeps the same rowstride. This is an effective way to "clip" the rgb_buffer for gdk.
       And we use gdk_draw_rgb_32_image instead of gdk_draw_pixbuf to avoid using Alpha channel */
    gdk_draw_rgb_32_image (widget->window, widget->style->white_gc,
        x, y,
        MIN (width - x, event->area.width), MIN (height - y, event->area.height),
        GDK_RGB_DITHER_MAX,
        gdk_pixbuf_get_pixels (buffer) + ((y * width + x) * 4),
        width * 4);
    return TRUE;
}

static gboolean
grdc_plug_vnc_on_configure (GtkWidget *widget, GdkEventConfigure *event, GrdcPlugVnc *gp_vnc)
{
    /* We do a delayed reallocating to improve performance */
    if (gp_vnc->scale_handler == 0)
    {
        gp_vnc->scale_handler = g_timeout_add (1000, (GSourceFunc) grdc_plug_vnc_update_scale_buffer, gp_vnc);
    }
    return FALSE;
}

static void
grdc_plug_vnc_class_init (GrdcPlugVncClass *klass)
{
    klass->parent_class.open_connection = grdc_plug_vnc_open_connection;
    klass->parent_class.close_connection = grdc_plug_vnc_close_connection;
    klass->parent_class.query_feature = grdc_plug_vnc_query_feature;
    klass->parent_class.call_feature = grdc_plug_vnc_call_feature;
}

static void
grdc_plug_vnc_destroy (GtkWidget *widget, gpointer data)
{
}

static void
grdc_plug_vnc_init (GrdcPlugVnc *gp_vnc)
{
    gp_vnc->drawing_area = gtk_drawing_area_new ();
    gtk_widget_show (gp_vnc->drawing_area);
    gtk_container_add (GTK_CONTAINER (gp_vnc), gp_vnc->drawing_area);

    gtk_widget_add_events (gp_vnc->drawing_area, GDK_POINTER_MOTION_MASK | GDK_BUTTON_PRESS_MASK
        | GDK_BUTTON_RELEASE_MASK | GDK_KEY_PRESS_MASK | GDK_KEY_RELEASE_MASK);
    GTK_WIDGET_SET_FLAGS (gp_vnc->drawing_area, GTK_CAN_FOCUS);

    g_signal_connect (G_OBJECT (gp_vnc->drawing_area), "expose_event",
        G_CALLBACK (grdc_plug_vnc_on_expose), gp_vnc);
    g_signal_connect (G_OBJECT (gp_vnc->drawing_area), "configure_event",
        G_CALLBACK (grdc_plug_vnc_on_configure), gp_vnc);
    g_signal_connect (G_OBJECT (gp_vnc), "destroy", G_CALLBACK (grdc_plug_vnc_destroy), NULL);

    gp_vnc->connected = FALSE;
    gp_vnc->running = FALSE;
    gp_vnc->auth_called = FALSE;
    gp_vnc->closed = FALSE;
    gp_vnc->rgb_buffer = NULL;
    gp_vnc->vnc_buffer = NULL;
    gp_vnc->scale_buffer = NULL;
    gp_vnc->scale_width = 0;
    gp_vnc->scale_height = 0;
    gp_vnc->scale_handler = 0;
    gp_vnc->clipboard_handler = 0;
    g_get_current_time (&gp_vnc->clipboard_timer);
    gp_vnc->client = NULL;
    gp_vnc->listen_sock = -1;
    gp_vnc->button_mask = 0;
    gp_vnc->pressed_keys = g_ptr_array_new ();
    gp_vnc->thread = 0;
}

GtkWidget*
grdc_plug_vnc_new (gboolean scale)
{
    GrdcPlugVnc *gp_vnc;

    gp_vnc = GRDC_PLUG_VNC (g_object_new (GRDC_TYPE_PLUG_VNC, NULL));
    GRDC_PLUG (gp_vnc)->scale = scale;

    return GTK_WIDGET (gp_vnc);
}

