/*
 * Grdc - GTK+/Gnome Remote Desktop Client
 * Copyright (C) 2009 - Vic Lee 
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, 
 * Boston, MA 02111-1307, USA.
 */

#include "config.h"

#ifdef HAVE_LIBSSH

#include <gtk/gtk.h>
#include <glib/gi18n.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/types.h>
#include <libssh/libssh.h>
#include <pthread.h>
#ifdef HAVE_NETDB_H
#include <netdb.h>
#endif
#ifdef HAVE_ARPA_INET_H
#include <arpa/inet.h>
#endif
#ifdef HAVE_NETINET_IN_H
#include <netinet/in.h>
#endif
#ifdef HAVE_SYS_SOCKET_H
#include <sys/socket.h>
#endif
#ifdef HAVE_FCNTL_H
#include <fcntl.h>
#endif
#ifdef HAVE_ERRNO_H
#include <errno.h>
#endif
#include "grdcssh.h"

static const gchar *common_identities[] = {
    ".ssh/id_rsa",
    ".ssh/id_dsa",
    ".ssh/identity",
    NULL
};

gchar*
grdc_ssh_identity_path (const gchar *id)
{
    if (id == NULL) return NULL;
    if (id[0] == '/') return g_strdup (id);
    return g_strdup_printf ("%s/%s", g_get_home_dir (), id);
}

gchar*
grdc_ssh_find_identity (void)
{
    gchar *path;
    gint i;

    for (i = 0; common_identities[i]; i++)
    {
        path = grdc_ssh_identity_path (common_identities[i]);
        if (g_file_test (path, G_FILE_TEST_IS_REGULAR | G_FILE_TEST_EXISTS))
        {
            g_free (path);
            return g_strdup (common_identities[i]);
        }
        g_free (path);
    }
    return NULL;
}

static void
grdc_ssh_tunnel_set_error (GrdcSSHTunnel *tunnel, const gchar *fmt)
{
    const gchar *err;

    err = ssh_get_error (tunnel->session);
    tunnel->error = g_strdup_printf (fmt, err);
}

GrdcSSHTunnel*
grdc_ssh_tunnel_new (const gchar *server, const gchar *user)
{
    GrdcSSHTunnel *tunnel;
    SSH_OPTIONS *options;
    gchar *server_name, *ptr;
    guint port;

    tunnel = g_new (GrdcSSHTunnel, 1);
    tunnel->session = NULL;
    tunnel->channel = NULL;
    tunnel->authenticated = FALSE;
    tunnel->tunnel_server_sock = -1;
    tunnel->tunnel_client_sock = -1;
    tunnel->thread = 0;
    tunnel->running = FALSE;
    tunnel->dest_host = NULL;
    tunnel->dest_port = 0;
    tunnel->buffer = NULL;
    tunnel->buffer_len = 0;
    tunnel->error = NULL;

    /* Parse the address and port */
    server_name = g_strdup (server);
    ptr = g_strrstr (server_name, ":");
    if (ptr != NULL)
    {
        *ptr++ =  '\0';
        port = (guint) atoi (ptr);
    }
    else
    {
        port = 22;
    }

    options = ssh_options_new ();
    ssh_options_set_username (options, (gchar*) user);
    ssh_options_set_host (options, server_name);
    ssh_options_set_port (options, port);

    g_free (server_name);

    /* Now init & startup the SSH session */    
    tunnel->session = ssh_new ();
    ssh_set_options (tunnel->session, options);
    if (ssh_connect (tunnel->session))
    {
        grdc_ssh_tunnel_set_error (tunnel, _("Failed to startup SSH session: %s"));
    }

    /* Try the "none" authentication */
    if (ssh_userauth_none (tunnel->session, NULL) == SSH_AUTH_SUCCESS)
    {
        tunnel->authenticated = TRUE;
    }

    return tunnel;
}

gboolean
grdc_ssh_tunnel_auth_password (GrdcSSHTunnel* tunnel, const gchar *password)
{
    gint ret;

    if (tunnel->authenticated) return TRUE;

    ret = ssh_userauth_password (tunnel->session, NULL, (gchar*) password);
    if (ret != SSH_AUTH_SUCCESS)
    {
        grdc_ssh_tunnel_set_error (tunnel, _("SSH password authentication failed: %s"));
        return FALSE;
    }

    tunnel->authenticated = TRUE;
    return TRUE;
}

gint
grdc_ssh_tunnel_auth_pubkey (GrdcSSHTunnel* tunnel, const gchar *pubkeyfile, 
    const gchar *privkeyfile, const gchar *passphrase)
{
    gint ret;
    STRING *pubkey;
    PRIVATE_KEY *privkey;
    gint keytype;

    if (tunnel->authenticated) return 1;

    pubkey = publickey_from_file (tunnel->session, (gchar*) pubkeyfile, &keytype);
    if (pubkey == NULL)
    {
        grdc_ssh_tunnel_set_error (tunnel, _("SSH public key authentication failed: %s"));
        return 0;
    }

    privkey = privatekey_from_file (tunnel->session, (gchar*) privkeyfile, keytype, (gchar*) passphrase);
    if (privkey == NULL)
    {
        if (passphrase == NULL || passphrase[0] == '\0') return -1;

        grdc_ssh_tunnel_set_error (tunnel, _("SSH public key authentication failed: %s"));
        /*string_free (pubkey);*/
        return 0;
    }

    ret = ssh_userauth_pubkey (tunnel->session, NULL, pubkey, privkey);
    /*string_free (pubkey);*/
    private_key_free (privkey);

    if (ret != SSH_AUTH_SUCCESS)
    {
        grdc_ssh_tunnel_set_error (tunnel, _("SSH public key authentication failed: %s"));
        return 0;
    }

    tunnel->authenticated = TRUE;
    return 1;
}

static gpointer
grdc_ssh_tunnel_main_thread (gpointer data)
{
    GrdcSSHTunnel *tunnel = (GrdcSSHTunnel*) data;
    gint flags;
    gchar *ptr;
    ssize_t len, lenw;
    fd_set set;
    struct timeval timeout;
    CHANNEL *ch[2], *chout[2];
    gint ret;
     
    pthread_setcancelstate (PTHREAD_CANCEL_ENABLE, NULL);
    pthread_setcanceltype (PTHREAD_CANCEL_ASYNCHRONOUS, NULL);

    /* Accept a local connection */
    tunnel->tunnel_client_sock = accept (tunnel->tunnel_server_sock, NULL, NULL);
    if (tunnel->tunnel_client_sock < 0)
    {
        tunnel->error = g_strdup ("Failed to accept local socket");
        tunnel->thread = 0;
        return NULL;
    }
    close (tunnel->tunnel_server_sock);
    tunnel->tunnel_server_sock = -1;

    if ((flags = fcntl (tunnel->tunnel_client_sock, F_GETFL, 0)) < 0) 
    { 
        tunnel->error = g_strdup ("fcntl(1) failed");
        tunnel->thread = 0;
        return NULL;
    } 

    if (fcntl (tunnel->tunnel_client_sock, F_SETFL, flags | O_NONBLOCK) < 0) 
    { 
        tunnel->error = g_strdup ("fcntl(2) failed");
        tunnel->thread = 0;
        return NULL;
    } 

    /* Request the SSH server to connect to the destination */
    tunnel->channel = channel_new (tunnel->session);
    if (channel_open_forward (tunnel->channel, tunnel->dest_host, tunnel->dest_port, "127.0.0.1", 0) != SSH_OK)
    {
        grdc_ssh_tunnel_set_error (tunnel, _("Failed to connect to the SSH tunnel destination: %s"));
        tunnel->thread = 0;
        return NULL;
    }

    tunnel->buffer_len = 10240;
    tunnel->buffer = g_malloc (tunnel->buffer_len);

    ch[0] = tunnel->channel;
    ch[1] = NULL;

    /* Start the tunnel data transmittion */
    while (tunnel->running)
    {
        timeout.tv_sec = 1;
        timeout.tv_usec = 0;

        FD_ZERO (&set);
        FD_SET (tunnel->tunnel_client_sock, &set);

        ret = ssh_select (ch, chout, FD_SETSIZE, &set, &timeout);
        if (!tunnel->running) break;
        if (ret == SSH_EINTR) continue;
        if (ret == -1) break;

        while (tunnel->running &&
            (len = read (tunnel->tunnel_client_sock, tunnel->buffer, tunnel->buffer_len)) > 0)
        {
            for (ptr = tunnel->buffer, lenw = 0; len > 0; len -= lenw, ptr += lenw)
            {
                lenw = channel_write (tunnel->channel, (char*) ptr, len);
                if (lenw <= 0)
                {
                    tunnel->running = FALSE;
                    break;
                }
            }
        }
        if (!tunnel->running || len == 0) break;

        len = channel_poll (tunnel->channel, 0);
        if (len == SSH_ERROR) break;
        if (len > 0)
        {
            if (len > tunnel->buffer_len)
            {
                tunnel->buffer_len = len;
                tunnel->buffer = (gchar*) g_realloc (tunnel->buffer, tunnel->buffer_len);
            }
            len = channel_read_nonblocking (tunnel->channel, tunnel->buffer, len, 0);
            if (len <= 0) break;

            for (ptr = tunnel->buffer, lenw = 0; len > 0; len -= lenw, ptr += lenw)
            {
                lenw = write (tunnel->tunnel_client_sock, ptr, len);
                if (lenw < 0 && tunnel->running)
                {
                    lenw = 0;
                    continue;
                }
                if (lenw <= 0)
                {
                    tunnel->running = FALSE;
                    break;
                }
            }
        }
        if (len == SSH_ERROR) break;
    }

    tunnel->thread = 0;
    return NULL;
}

gboolean
grdc_ssh_tunnel_start (GrdcSSHTunnel* tunnel, const gchar *dest, gint local_port)
{
    gchar *ptr;
    gint sock;
    gint sockopt = 1;
    struct sockaddr_in sin;

    tunnel->dest_host = g_strdup (dest);
    ptr = g_strrstr (tunnel->dest_host, ":");
    if (ptr != NULL)
    {
        *ptr++ =  '\0';
        tunnel->dest_port = atoi (ptr);
    }
    else
    {
        tunnel->error = g_strdup ("Destination port has not been assigned");
        return FALSE;
    }

    /* Create the server socket that listens on the local port */
    sock = socket (AF_INET, SOCK_STREAM, 0);
    if (sock < 0)
    {
        tunnel->error = g_strdup ("Failed to create socket.");
        return FALSE;
    }
    setsockopt (sock, SOL_SOCKET, SO_REUSEADDR, &sockopt, sizeof (sockopt));

    sin.sin_family = AF_INET;
    sin.sin_port = htons (local_port);
    sin.sin_addr.s_addr = inet_addr("127.0.0.1");

    if (bind (sock, (struct sockaddr *) &sin, sizeof(sin)))
    {
        tunnel->error = g_strdup ("Failed to bind on local port.");
        close (sock);
        return FALSE;
    }

    if (listen (sock, 1))
    {
        tunnel->error = g_strdup ("Failed to listen on local port.");
        close (sock);
        return FALSE;
    }

    tunnel->tunnel_server_sock = sock;
    tunnel->running = TRUE;

    if (pthread_create (&tunnel->thread, NULL, grdc_ssh_tunnel_main_thread, tunnel))
    {
        tunnel->error = g_strdup ("Failed to initialize pthread.");
        tunnel->thread = 0;
        return FALSE;
    }
    return TRUE;
}

gboolean
grdc_ssh_tunnel_terminated (GrdcSSHTunnel* tunnel)
{
    return (tunnel->thread == 0);
}

void
grdc_ssh_tunnel_free (GrdcSSHTunnel* tunnel)
{
    if (tunnel->thread != 0)
    {
        tunnel->running = FALSE;
        pthread_cancel (tunnel->thread);
        pthread_join (tunnel->thread, NULL);
    }
    if (tunnel->tunnel_client_sock >= 0) close (tunnel->tunnel_client_sock);
    if (tunnel->tunnel_server_sock >= 0) close (tunnel->tunnel_server_sock);
    if (tunnel->channel)
    {
        channel_close (tunnel->channel);
        channel_free (tunnel->channel);
        tunnel->channel = NULL;
    }
    if (tunnel->session)
    {
        ssh_disconnect (tunnel->session);
        tunnel->session = NULL;
    }
    if (tunnel->buffer)
    {
        g_free (tunnel->buffer);
        tunnel->buffer = NULL;
    }
    g_free (tunnel->dest_host);
    g_free (tunnel->error);
    g_free (tunnel);
}

#endif /* HAVE_LIBSSH */

